//============================================================================
// CAN^XN
//
// (c)2021 Micon-Giken
//============================================================================
#include "kernel.h"
#include "task_can.h"

//---------------------------------------------------------------------
// `.
//---------------------------------------------------------------------
//void CAN_main(void);				// pPbg.

typedef struct {
    int		id;
    int		dlc;
    uchar	data[8];
} CAN_BUFFER;

typedef struct {
	uint	id2:16;
	uint	id1:11;
	uint	act:3;
	uint	type:2;
} CAN_FILTER;

typedef enum {
	CAN_FILACT_DISABLE,
	CAN_FILACT_STF0M,
	CAN_FILACT_STF1M,
	CAN_FILACT_REJECT,
	CAN_FILACT_PRIORITY,
	CAN_FILACT_PRIF0M,
	CAN_FILACT_PRIF1M,
	CAN_FILACT_STRXBUF,
} CAN_FILACT;

typedef enum {
	CAN_FILTYPE_RANGE,
	CAN_FILTYPE_DUAL,
	CAN_FILTYPE_CLASSIC,
} CAN_FILTYPE;

//---------------------------------------------------------------------
// f[^.
//---------------------------------------------------------------------
static CAN_FILTER CAN0_id_filt[CAN_FILT_SIZE] = {	// MtB^.
//	{0x1ff, 0x000, CAN_FILACT_STF0M, CAN_FILTYPE_CLASSIC},
	{0x1ff, 0x000, CAN_FILACT_STF0M, CAN_FILTYPE_RANGE},
};
static volatile CAN_BUFFER CAN0_rx_fifo[CAN_RECV_SIZE];  // MFIFO.
static volatile CAN_BUFFER CAN0_tx_fifo[CAN_SEND_SIZE];  // MFIFO.

static CAN_FILTER CAN1_id_filt[CAN_FILT_SIZE] = {	// MtB^.
//	{0x1ff, 0x000, CAN_FILACT_STF0M, CAN_FILTYPE_CLASSIC},
	{0x1ff, 0x000, CAN_FILACT_STF0M, CAN_FILTYPE_RANGE},
};
static volatile CAN_BUFFER CAN1_rx_fifo[CAN_RECV_SIZE];  // MFIFO.
static volatile CAN_BUFFER CAN1_tx_fifo[CAN_SEND_SIZE];  // MFIFO.

static	int	flag;				// MǗtO.

//---------------------------------------------------------------------
// hCoAPI
//---------------------------------------------------------------------
// CANhCo.
void CAN_init(void) {
	uint	wk;
	
    //GCLK,oXNbNݒ.
	_reg(GCLK, PCHCTRL)[CAN0_PCH] =
			_val(GCLK, PCHCTRL, GEN, CAN_GCLK) |
			_val(GCLK, PCHCTRL, CHEN, 1) ;
	CAN0_AHB_EN();										//oXNbNJn.
	_reg(GCLK, PCHCTRL)[CAN1_PCH] =
			_val(GCLK, PCHCTRL, GEN, CAN_GCLK) |
			_val(GCLK, PCHCTRL, CHEN, 1) ;
	CAN1_AHB_EN();										//oXNbNJn.

    //|[gݒ.
    PORT_peri(CAN0_TX_PIN, CAN_MAPPING);
    PORT_peri(CAN0_RX_PIN, CAN_MAPPING);
    PORT_peri(CAN1_TX_PIN, CAN_MAPPING);
    PORT_peri(CAN1_RX_PIN, CAN_MAPPING);

    //CANݒ.
	_setCAN(CAN0, CCCR, INIT, 1);
    while (!_bitCAN(CAN0, CCCR, INIT));
	_setCAN(CAN1, CCCR, INIT, 1);
    while (!_bitCAN(CAN1, CCCR, INIT));
	_setCAN(CAN0, CCCR, CCE, 1);
    while (!_bitCAN(CAN0, CCCR, CCE));
	_setCAN(CAN1, CCCR, CCE, 1);
    while (!_bitCAN(CAN1, CCCR, CCE)) {}			// {}̓[jO΍.
	
	wk =
			_valCAN(CAN0, CCCR, INIT, 1) |
			_valCAN(CAN0, CCCR, CCE , 1) |
			_valCAN(CAN0, CCCR, TXP , 1) |
			_valCAN(CAN0, CCCR, DAR , 0) ;			// đ.
//			_valCAN(CAN0, CCCR, DAR , 1) ;			// đ֎~.
	_regCAN(CAN0, CCCR) = wk;
	_regCAN(CAN1, CCCR) = wk;
	
//	_regCAN(CAN0, MRCFG) = CAN_MRCFG_QOS_MEDIUM;	// Sensitive latency(default)
//	_regCAN(CAN1, MRCFG) = CAN_MRCFG_QOS_MEDIUM;	// Sensitive latency(default)
//	_regCAN(CAN0, DBTP ) = 0x00000a33;				// 500kbps(default)
//	_regCAN(CAN0, NBTP ) = 0x06000a03;				// 500kbps(default)
//	_regCAN(CAN1, DBTP ) = 0x00000a33;				// 500kbps(default)
//	_regCAN(CAN1, NBTP ) = 0x06000a03;				// 500kbps(default)

	//tB^AMFIFOAMFIFO̐ݒ.
	_regCAN(CAN0, SIDFC) =
			_valCAN(CAN0, SIDFC, LSS, CAN_FILT_SIZE) |
			((uint)CAN0_id_filt & 0x0ffff);
	_regCAN(CAN1, SIDFC) =
			_valCAN(CAN1, SIDFC, LSS, CAN_FILT_SIZE) |
			((uint)CAN1_id_filt & 0x0ffff);
	_regCAN(CAN0, RXF0C) =
			_valCAN(CAN0, RXF0C, F0OM, 0) |
			_valCAN(CAN0, RXF0C, F0S , CAN_RECV_SIZE) |
			((uint)CAN0_rx_fifo & 0x0ffff);
	_regCAN(CAN1, RXF0C) =
			_valCAN(CAN1, RXF0C, F0OM, 0) |
			_valCAN(CAN1, RXF0C, F0S , CAN_RECV_SIZE) |
			((uint)CAN1_rx_fifo & 0x0ffff);
	_regCAN(CAN0, TXBC) =
			_valCAN(CAN0, TXBC, TFQM, 0) |
			_valCAN(CAN0, TXBC, TFQS, CAN_SEND_SIZE) |
			((uint)CAN0_tx_fifo & 0x0ffff);
	_regCAN(CAN1, TXBC) =
			_valCAN(CAN1, TXBC, TFQM, 0) |
			_valCAN(CAN1, TXBC, TFQS, CAN_SEND_SIZE) |
			((uint)CAN1_tx_fifo & 0x0ffff);

	//݂͕sgp.
    //MJn.
	wk &= ~3;
	_regCAN(CAN0, CCCR) = wk;
	_regCAN(CAN1, CCCR) = wk;
	
//	start_task(CAN_taskid, CAN_main);
}

//---------------------------------------------------------------------
// CANpPbgM.
int	CAN_send(CAN_PACKET *pp) {
	int		n, ret, index;
	
	if (pp->port == 0) {
		ret = _getCAN(CAN0, TXFQS, TFFL);
		if (ret > 0) {
			index = _getCAN(CAN0, TXFQS, TFQPI);
			CAN0_tx_fifo[index].id = pp->id << 18;
			CAN0_tx_fifo[index].dlc = pp->dlc << 16;
			for (n = 0; n < pp->dlc; n++) {
				CAN0_tx_fifo[index].data[n] = pp->data[n];
			}
			_regCAN(CAN0, TXBAR) = 1 << index;
			while(index == _getCAN(CAN0, TXFQS, TFQPI));
		}
	}
	else {
		ret = _getCAN(CAN1, TXFQS, TFFL);
		if (ret > 0) {
			index = _getCAN(CAN1, TXFQS, TFQPI);
			CAN1_tx_fifo[index].id = pp->id << 18;
			CAN1_tx_fifo[index].dlc = pp->dlc << 16;
			for (n = 0; n < pp->dlc; n++) {
				CAN1_tx_fifo[index].data[n] = pp->data[n];
			}
			_regCAN(CAN1, TXBAR) = 1 << index;
			while(index == _getCAN(CAN1, TXFQS, TFQPI));
		}
	}
	ret--;
	return ret;
}

//---------------------------------------------------------------------
// CANpPbgM.
int CAN_receive(CAN_PACKET *pp) {
	int n, ret, index;
	
	for (n = 0; n < 2; n++) {
		if (flag == 0) {
			flag ^= 1;
			ret = _getCAN(CAN0, RXF0S, F0FL);
			if (ret > 0) {
				index = _getCAN(CAN0, RXF0S, F0GI);
				pp->port = 0;
				pp->id  = (ushort)(CAN0_rx_fifo[index].id >> 18);
				pp->dlc = (uchar)(CAN0_rx_fifo[index].dlc >> 16) & 0x0f;
				for (n = 0; n < 8; n++) {
					if (n < pp->dlc) pp->data[n] = CAN0_rx_fifo[index].data[n];
					else			 pp->data[n] = 0;
				}
				_regCAN(CAN0, RXF0A) = index;
				while (index == _getCAN(CAN0, RXF0S, F0GI));
				ret--;
				ret += _getCAN(CAN1, RXF0S, F0FL);
				return ret;
			}
		}
		else {
			flag ^= 1;
			ret = _getCAN(CAN1, RXF0S, F0FL);
			if (ret > 0) {
				index = _getCAN(CAN1, RXF0S, F0GI);
				pp->port = 1;
				pp->id  = (ushort)(CAN1_rx_fifo[index].id >> 18);
				pp->dlc = (uchar)(CAN1_rx_fifo[index].dlc >> 16) & 0x0f;
				for (n = 0; n < 8; n++) {
					if (n < pp->dlc) pp->data[n] = CAN1_rx_fifo[index].data[n];
					else			 pp->data[n] = 0;
				}
				_regCAN(CAN1, RXF0A) = index;
				while (index == _getCAN(CAN1, RXF0S, F0GI));
				ret--;
				ret += _getCAN(CAN0, RXF0S, F0FL);
				return ret;
			}
		}
	}
	return EOF;
}

//---------------------------------------------------------------------
// hCo^XN.
//---------------------------------------------------------------------
#if 0
void CAN_main(void) {				// pPbg.
    sleep_task();
}
#endif

//---------------------------------------------------------------------
// 荞݊֐.
//---------------------------------------------------------------------

//============================================================================
