//============================================================================
// LCD^XN
//
// (c)2021-2022 Micon-Giken
//============================================================================
#include "kernel.h"
#include "task_lcd.h"
#include "Font.h"
#include <stdio.h>

//============================================================================
#if TARGET_LCD == LCD_TYPE_SG12864
//============================================================================

//============================================================================
// `.
//============================================================================
#define LCD_RESET_TIME		1		//LCDZbg.
#define LCD_INIT_LEN		2		//LCDf[^.
#define LCD_PULSE_TIME		1		//LCD݃pX.
#define LCD_WAIT_TIME		6		//LCD݌҂.
#define LCD_CURSOR_TIME		500000	// J[\uNon/off.
#define	LCD_REVERSE_ON		0x17	// o[XONR}h.
#define	LCD_REVERSE_OFF		0x16	// o[XOFFR}h.
#define	LCD_POSITION		0x18	// J[\ʒuR}h.

//============================================================================
// vg^Cv.
//============================================================================
void LCD_init1(void);					// LCD^XN1.
void LCD_init2(void);					// LCD^XN2.
void LCD_main(void);					// \^XN(又).

void LCD_wake(void);					// stdiopAPI.
void LCD_write_data(char c);			// LCD݃f[^o.
void LCD_write_pulse(void);			// LCDf[^݃pXo.
void LCD_cursor(char c);				// J[\\.
//	@: char c;							w(0:, 1:\)
void LCD_control(char c);				// e퐧R[h(\b,\t,\n)̏.
//	@: char c;							R[h.
void LCD_font(char *p, schar i);			// tHg.
//	@: char *p;							tHgEf[^ւ̃|C^.
//			char i;							1hbg󔒕t.
//								-1ȂOɕt,1Ȃɕt,0ȂtȂ.

//============================================================================
// f[^.
//============================================================================
static uchar LCD_buff[LCD_BUFF_SIZE];		// LCDOEobt@.
#if   STDIO_VERSION == 1
static FILE LCD_file = {					// LCD FILE
	LCD_buff, LCD_buff, LCD_buff, LCD_buff,
	LCD_buff + (LCD_BUFF_SIZE - 1), 0,0,0,0,
	LCD_wake, LCD_close
};
#elif STDIO_VERSION == 2
static FILE LCD_file = {					// LCD FILE
	LCD_buff, 0,0,0, LCD_BUFF_SIZE - 1, 0,0,0,0,
	LCD_wake, LCD_close
};
#endif

static int	LCD_v;					//JE^.
static int	LCD_h;					//JE^.
static ulong LCD_cursor_time;		//J[\uNJn.
static char	 LCD_cursor_data;		//J[\f[^.
static char	 LCD_reverse;			//o[Xf[^.
static const char LCD_init_data[LCD_INIT_LEN] = {		//LCDf[^.
	0xc0, 0x3f
};
#define wk	 LCD_reverse			//p[N.

//============================================================================
// API֐.
//============================================================================
// LCDI[v֐.
FILE* LCD_open(const char *name, const char *mode) {
	if (*mode == 'w') {
		if (LCD_file.rw_mode)	return NULL;
#if   STDIO_VERSION == 1
		LCD_file.wp = LCD_buff;
		LCD_file.rp = LCD_buff;
		LCD_file.cp = LCD_buff;
#elif STDIO_VERSION == 2
		LCD_file.wp = 0;
		LCD_file.rp = 0;
		LCD_file.cp = 0;
#endif
		LCD_file.rw_mode = mode[0];
		LCD_file.binary = mode[1];
		LCD_file.code = 0;
		LCD_file.type = 0;
		return &LCD_file;
	}
	return NULL;
}

//---------------------------------------------------------------------
// LCDN[Y֐.
int LCD_close(void *fp) {
	if (fp == &LCD_file) {
		LCD_file.rw_mode = '\0';
		LCD_file.wp = 0;
		LCD_file.rp = 0;
		LCD_file.cp = 0;
		return 0;
	}
	return EOF;
}

//---------------------------------------------------------------------
// LCDhCoE^XN.
void LCD_init() {
	
	//|[gݒ.
	PORT_mode_out(LCD_RST, 0);
	PORT_mode_out(LCD_CS2, 0);
	PORT_mode_out(LCD_CS1, 0);
	PORT_mode_out(LCD_D7, 0);
	PORT_mode_out(LCD_D6, 0);
	PORT_mode_out(LCD_D5, 0);
	PORT_mode_out(LCD_D4, 0);
	PORT_mode_out(LCD_D3, 0);
	PORT_mode_out(LCD_D2, 0);
	PORT_mode_out(LCD_D1, 0);
	PORT_mode_out(LCD_D0, 0);
	PORT_mode_out(LCD_E, 0);
	PORT_mode_out(LCD_DI, 0);
	
	start_task(LCD_taskid, LCD_init1);
}

//---------------------------------------------------------------------
// MJn (ԐAPI)
void LCD_wake(void) {
	__DI();
	wakeup_task(LCD_taskid);
	__EI();
	return;
}

//============================================================================
// ^XNA又.
//============================================================================
//1.
void LCD_init1(void) {
	ulong time;
	int	i;
	
	//Zbg.
	time = KERNEL_time();
	while (KERNEL_time() - time < LCD_RESET_TIME);
	PORT_outset(LCD_RST);
	
	//lLCDW[ɏ.
	PORT_outset(LCD_CS1);
	PORT_outset(LCD_CS2);
	for (i = 0; i < LCD_INIT_LEN; i++) {
		LCD_write_data(LCD_init_data[i]);
		LCD_write_pulse();
	}

	wk = 0;
	LCD_h = 0;
	next_task(LCD_init2);
}
//2.
void LCD_init2(void) {
	int j;
	//
	PORT_outclr(LCD_DI);			//R}hݎw.
	LCD_write_data(wk + 0xB8);		//y[WAhXݒ.
	LCD_write_pulse();
	LCD_write_data(LCD_h + 0x40);	//JAhXݒ.
	LCD_write_pulse();

	// NAEf[^](1x2)
	PORT_outset(LCD_DI);			//f[^ݎw.
	LCD_write_data(0);
	for (j = 0; j < 8; j++) {
		LCD_write_pulse();
	}
	
	//I
	LCD_h += 8;
	if (LCD_h < LCD_H_MAX * 8 / 2) {
		retry_task();
	}
	else {
		LCD_h = 0;
		if (++wk < LCD_V_MAX * 2) {
			retry_task();
		}
	}
	PORT_outclr(LCD_CS1);
	PORT_outclr(LCD_CS2);
	LCD_v = 0;
	LCD_h = 0;
	LCD_cursor_time = KERNEL_time();
	LCD_cursor_data = 0;
	LCD_reverse = 0;
	next_task(LCD_main);
}

//---------------------------------------------------------------------
//又.
void LCD_main(void) {
	int c;
	uchar d;
	char *p;

	// J[\EuN
	if ((c = KERNEL_time() - LCD_cursor_time) > LCD_CURSOR_TIME) {
		LCD_cursor_time += LCD_CURSOR_TIME;
		LCD_cursor_data ^= 0xff;
		LCD_cursor(1);
	}
	
	//sp.
	if (LCD_file.code == '\n') {
		LCD_control('\n');
		retry_task();
	}
	
	/***************************/
	/* \&R[h  */
	/***************************/
	if ( (c = getc(&LCD_file) ) == EOF) {
		delay_task(1);
	}
	d = (uchar) c;
	if (!LCD_file.code) {
		/* 1oCgڏ */
		if (d >= 0x80)
			LCD_file.code = d;					/* 2oCgR[h1oCgڕۑ */
		else if (d >= 0x20)	{					/* p\ */
			//LCD_cursor(0);						/* J[\폜 */	/* {J[\̏ꍇ͕݂ŏ */
			LCD_font( FONT_ascii(d), -1);			/* ptHgWJ */
			LCD_cursor(1);							/* J[\\ */
		}
		else if (d == LCD_POSITION)
			LCD_file.code = d;					/* J[\ړR}hۑ */
		else LCD_control((char)d);				/* ̑R[h */
	}
	else if (LCD_file.code > 0x7f) {
		/* 2oCg\ */
		c = ((int) LCD_file.code << 8) + d;		/* 2oCgR[h̐ */
		p = FONT_kanji( (unsigned short) c);		/* 2oCgR[hSptHgʒu擾 */
		//LCD_cursor(0);						/* J[\폜 */	/* {J[\̏ꍇ͕݂ŏ */
		LCD_font(p, -1);						/* Sp̑O */
		LCD_font(p + 14, 1);					/* Sp̌㔼 */
		LCD_cursor(1);							/* J[\\ */
		LCD_file.code = 0;						/* 2oCgI */
	}
	else if (LCD_file.code == LCD_POSITION) {
		/* J[\ړ(p[^1oCgڂ̕ۑ) */
		LCD_file.type = d;
		LCD_file.code++;
	}
	else if (LCD_file.code == LCD_POSITION + 1) {
		/* J[\ړ */
		LCD_cursor(0);							/* J[\폜 */
		if (!(LCD_file.type & 2)) {				/* Έʒuړ̏ꍇ͌ʒuNA */
			LCD_v = 0;
			LCD_h = 0;
		}
		LCD_v += (schar)(LCD_file.type >> 2);
		LCD_h += (schar)(d >> 1);
		while (LCD_v >= LCD_V_MAX) LCD_v -= LCD_V_MAX;
		while (LCD_h >= LCD_H_MAX) LCD_h -= LCD_H_MAX;
		LCD_cursor(1);							/* J[\\ */
		LCD_file.code = 0;						/* ړR}hI */
	}
	retry_task();
}

//============================================================================
// [J֐.
//============================================================================
// LCD݃f[^o.
void LCD_write_data(char c) {
	
	//f[^oXݒ.
	PORT_outbit(LCD_D7, c & 0x80);
	PORT_outbit(LCD_D6, c & 0x40);
	PORT_outbit(LCD_D5, c & 0x20);
	PORT_outbit(LCD_D4, c & 0x10);
	PORT_outbit(LCD_D3, c & 0x08);
	PORT_outbit(LCD_D2, c & 0x04);
	PORT_outbit(LCD_D1, c & 0x02);
	PORT_outbit(LCD_D0, c & 0x01);

	return;
}

void LCD_write_pulse(void) {
	ulong time;
	
	//݃pXo.
	time = KERNEL_time();
	while (KERNEL_time() - time < LCD_PULSE_TIME);
	PORT_outset(LCD_E);
	time = KERNEL_time();
	while (KERNEL_time() - time < LCD_PULSE_TIME);
	PORT_outclr(LCD_E);
	time = KERNEL_time();
	while (KERNEL_time() - time < LCD_WAIT_TIME);
	return;
}

//---------------------------------------------------------------------
// e퐧R[h(\b,\t,\n)̏.
void LCD_control(char c) {
	char i;

	LCD_cursor(0);					/* J[\폜 */
	switch (c) {
		case '\b':
			/*****************************************
			 obNXy[X
			 *****************************************/
			if (--LCD_h < 0) {
				LCD_h = LCD_H_MAX - 1;
				if (--LCD_v < 0) LCD_v = LCD_V_MAX - 1;
			}
			LCD_font(0, 1);
			if (--LCD_h < 0) {
				LCD_h = LCD_H_MAX - 1;
				if (--LCD_v < 0) LCD_v = LCD_V_MAX - 1;
			}
			break;
		case '\t':
			/*****************************************
			 ^u(p4P)
			 *****************************************/
			if ((i = (char) (LCD_h & 3) )) {
				i = (char) (4 - i);
				while (i--) LCD_font(0, 1);
			}
			break;
		case '\n':
			/*****************************************
			 s
			 *****************************************/
			i = (char) (LCD_H_MAX - LCD_h);
			if (i > 1)	LCD_file.code = c;
			else		LCD_file.code = 0;
//			while (i--) {
				LCD_font(0, 1);
//			}
			break;
		case LCD_REVERSE_ON:  LCD_reverse = 0xff; break;
		case LCD_REVERSE_OFF: LCD_reverse = 0;    break;
		default:
			break;
	}
	LCD_cursor(1);				/* J[\\ */
	return;
}

//---------------------------------------------------------------------
// tHg.
void LCD_font(char *p, schar i) {
	uchar j, k, m, n;

	/* ݃CʒuAJݒ */
	if (LCD_h < 8)	PORT_outset(LCD_CS1);
	else			PORT_outset(LCD_CS2);

	m = (uchar) (0xB8 + LCD_v + LCD_v);		/* y[WAhX */
	n = (uchar) (LCD_h << 3);				/* JEAhX */
	if (n < 0x40) n += 0x40;

	for (k = 0; k < 2; k++) {
		/***** ݃AhXݒ *****/
		PORT_outclr(LCD_DI);				/* R}hݎw */
		LCD_write_data(m++);				/* y[WAhXݒ */
		LCD_write_pulse();
		LCD_write_data(n);					/* JAhXݒ */
		LCD_write_pulse();

		/***** tHgEf[^] *****/
		PORT_outset(LCD_DI);				/* f[^ݎw */
		LCD_write_data(LCD_reverse);		/* Oɋ󔒒ǉ */
		if (i < 0) {
			LCD_write_pulse();
		}
		
		for (j = 0; j < 7; j++) {			/* tHg] */
			if (p != NULL)	LCD_write_data((char)(*p++ ^ LCD_reverse));
			LCD_write_pulse();
		}

		LCD_write_data(LCD_reverse);		/* ɋ󔒒ǉ */
		if (i > 0) {
			for (j = 0; j < i; j++) {
				LCD_write_pulse();
			}
		}
	}

	/* LCD I */
	PORT_outclr(LCD_CS1);
	PORT_outclr(LCD_CS2);

	if (++LCD_h >= LCD_H_MAX) {
		LCD_h = 0;
		if (++LCD_v >= LCD_V_MAX) LCD_v = 0;
	}
	
	return;
}

//---------------------------------------------------------------------
// J[\\.
void LCD_cursor(char c) {
	uchar k, m, n; //, LCD_D;

	/* ݃CʒuAJݒ */
	if (LCD_h < 8)	PORT_outset(LCD_CS1);
	else			PORT_outset(LCD_CS2);

	if (c) c = LCD_cursor_data;					/* J[\Ef[^ */
	m = (uchar) (0xB8 + LCD_v + LCD_v);			/* y[WAhX */
	n = (uchar) (LCD_h << 3);					/* JEAhX */
	if (n < 0x40) n += 0x40;

	for (k = 0; k < 2; k++) {
		/***** ݃AhXݒ *****/
		PORT_outclr(LCD_DI);					/* R}hݎw */
		LCD_write_data(m++);					/* y[WAhXݒ */
		LCD_write_pulse();
		LCD_write_data(n);						/* JAhXݒ */
		LCD_write_pulse();

		/***** J[\] *****/
		PORT_outset(LCD_DI);					/* f[^ݎw */
		LCD_write_data(c);
		LCD_write_pulse();
	}

	/* LCD I */
	PORT_outclr(LCD_CS1);
	PORT_outclr(LCD_CS2);

	return;
}

//============================================================================
// nh.
//============================================================================

//============================================================================
#endif




//============================================================================
#if TARGET_LCD == LCD_TYPE_AQM1248
//============================================================================

//============================================================================
// `.
//============================================================================
#define LCD_CURSOR_TIME		500000	// J[\uNon/off.
#define	LCD_REVERSE_ON		0x17	// o[XONR}h.
#define	LCD_REVERSE_OFF		0x16	// o[XOFFR}h.
#define	LCD_POSITION		0x18	// J[\ʒuR}h.

//============================================================================
// vg^Cv.
//============================================================================
void LCD_init1(void);					// LCD^XN1.
void LCD_init2(void);					// LCD^XN2.
void LCD_init3(void);					// LCD^XN3.
void LCD_init4(void);					// LCD^XN4.
void LCD_main(void);					// \^XN(又).

void LCD_wake(void);					// stdiopAPI.
void LCD_write_data(char c);			// LCD݃f[^o.
void LCD_cursor(char c);				// J[\\.
//	@: char c;							w(0:, 1:\)
void LCD_control(char c);				// e퐧R[h(\b,\t,\n)̏.
//	@: char c;							R[h.
void LCD_font(char *p, schar i);			// tHg.
//	@: char *p;							tHgEf[^ւ̃|C^.
//			char i;							1hbg󔒕t.
//								-1ȂOɕt,1Ȃɕt,0ȂtȂ.

//============================================================================
// f[^.
//============================================================================
static uchar LCD_buff[LCD_BUFF_SIZE];		// LCDOEobt@.
#if   STDIO_VERSION == 1
static FILE LCD_file = {					// LCD FILE
	LCD_buff, LCD_buff, LCD_buff, LCD_buff,
	LCD_buff + (LCD_BUFF_SIZE - 1), 0,0,0,0,
	LCD_wake, LCD_close
};
#elif STDIO_VERSION == 2
static FILE LCD_file = {					// LCD FILE
	LCD_buff, 0,0,0, LCD_BUFF_SIZE - 1, 0,0,0,0,
	LCD_wake, LCD_close
};
#endif

static int	LCD_v;					//JE^.
static int	LCD_h;					//JE^.
static ulong LCD_cursor_time;		//J[\uNJn.
static char	 LCD_cursor_data;		//J[\f[^.
static char	 LCD_reverse;			//o[Xf[^.
#define wk	 LCD_reverse			//p[N.

//============================================================================
// API֐.
//============================================================================
// LCDI[v֐.
FILE* LCD_open(const char *name, const char *mode) {
	if (*mode == 'w') {
		if (LCD_file.rw_mode)	return NULL;
#if   STDIO_VERSION == 1
		LCD_file.wp = LCD_buff;
		LCD_file.rp = LCD_buff;
		LCD_file.cp = LCD_buff;
#elif STDIO_VERSION == 2
		LCD_file.wp = 0;
		LCD_file.rp = 0;
		LCD_file.cp = 0;
#endif
		LCD_file.rw_mode = mode[0];
		LCD_file.binary = mode[1];
		LCD_file.code = 0;
		LCD_file.type = 0;
		return &LCD_file;
	}
	return NULL;
}

//---------------------------------------------------------------------
// LCDN[Y֐.
int LCD_close(void *fp) {
	if (fp == &LCD_file) {
		LCD_file.rw_mode = '\0';
		LCD_file.wp = 0;
		LCD_file.rp = 0;
		LCD_file.cp = 0;
		return 0;
	}
	return EOF;
}

//---------------------------------------------------------------------
// LCDhCoE^XN.
void LCD_init() {
	//ӗpNbNAoXNbNݒ.
#if (TARGET_BOARD & SAMD21_ALL)
	_reg(GCLK, CLKCTRL) =
			_val(GCLK, CLKCTRL, ID, SPI_PCH) |
			_val(GCLK, CLKCTRL, GEN, SPI_GCLK) |
			_val(GCLK, CLKCTRL, CLKEN, 1) ;
#else
	_reg(GCLK, PCHCTRL)[SPI_PCH] =
			_val(GCLK, PCHCTRL, GEN, SPI_GCLK) |
			_val(GCLK, PCHCTRL, CHEN, 1) ;
#endif
	SPI_APB_EN();										//oXNbNJn.
	
	//|[gݒ.
	PORT_mode_out(LCD_RS, 0);
	PORT_mode_out(LCD_CS, 1);
	PORT_peri(LCD_DO, SPI_MAPPING);
	PORT_peri(LCD_SCK, SPI_MAPPING);

	//SPIݒ.
	#define SPI_BAUD	6000000
	#if (UART_FREQUENCY / 2 / SPI_BAUD) > 0
	_regSP(SPI_SP, BAUD) = UART_FREQUENCY / 2 / SPI_BAUD - 1;
	#else
	_regSP(SPI_SP, BAUD) = 0;
	#endif
	
	_regSP(SPI_SP, CTRLA) =
			_valSP(SPI_SP, CTRLA, DORD, 0)	   |		//MSB first.
			_valSP(SPI_SP, CTRLA, CPOL, 1)	   |		//SCK is Negative.
			_valSP(SPI_SP, CTRLA, CPHA, 1)	   |		//Falling change, Rising sample.
			_valSP(SPI_SP, CTRLA, FORM, 0)	   |		//SPI mode(without ADDR).
#if (TARGET_BOARD & SAMD21_ALL)
			_valSP(SPI_SP, CTRLA, DIPO, 0x02)  |		//M|[g(_~[).
			_valSP(SPI_SP, CTRLA, DOPO, 0x01)  |		//M|[g.
#else
			_valSP(SPI_SP, CTRLA, DIPO, 0x03)  |		//M|[g(_~[).
			_valSP(SPI_SP, CTRLA, DOPO, 0x00)  |		//M|[g.
#endif
			_valSP(SPI_SP, CTRLA, MODE, 0x03)  ;		//master[h.
	_setSP(SPI_SP, CTRLA, ENABLE, 1);					//Jn.

//	NVIC_set_priority(SPI_IRQ, 192);					//NVICDx192.
//	NVIC_enable_irq(SPI_IRQ);							//NVIC.
	_regSP(SPI_SP, INTENSET) = SPI_TX_END;				//M݃Cl[u.

	start_task(LCD_taskid, LCD_init1);
}

//---------------------------------------------------------------------
// MJn (ԐAPI)
void LCD_wake(void) {
	__DI();
	wakeup_task(LCD_taskid);
	__EI();
	return;
}

//============================================================================
// ^XNA又.
//============================================================================
//0-3.
void LCD_init1(void) {

	//lLCDW[ɏ.
	PORT_outclr(LCD_CS);
	LCD_write_data(0xAE);		//fBXvCOFF.
	LCD_write_data(0xA0);		//ADC = nomal.
	LCD_write_data(0xC8);		//Common out = reverse.
	LCD_write_data(0xA3);		//bias = 1/7.
	LCD_write_data(0x2C);		//p[1.

	break_delay_task(LCD_init2, 5);
	LCD_write_data(0x2E);		//p[2.

	break_delay_task(LCD_init3, 5);
	LCD_write_data(0x2F);		//p[3.

	LCD_write_data(0x23);		//RgXg1.
	LCD_write_data(0x81);		//RgXg2.
	LCD_write_data(0x1C);		//RgXg3.
	LCD_write_data(0xA4);		//display all point = normal.
	LCD_write_data(0x40);		//display start line = 0.
	LCD_write_data(0xA6);		//o[XȂ.
	LCD_write_data(0xAF);		//fBXvCON.
	PORT_outset(LCD_CS);

	wk = 0;
	next_task(LCD_init4);
}
//2.
void LCD_init4(void) {
	int j;
	//
	PORT_outclr(LCD_CS);
	PORT_outclr(LCD_RS);			//R}hݎw.
	LCD_write_data(wk + 0xB0);		//y[WAhXݒ.
	LCD_write_data(0x10);			//JAhXʐݒ.
	LCD_write_data(0x00);			//JAhXʐݒ.

	// NAEf[^](1s)
	PORT_outset(LCD_RS);			//f[^ݎw.
	for (j = 0; j < LCD_H_MAX * 8; j++) {
		LCD_write_data(0x00);
	}
	PORT_outset(LCD_CS);
	
	//I
	if (++wk < LCD_V_MAX * 2) {
		retry_task();
	}
	LCD_v = 0;
	LCD_h = 0;
	LCD_cursor_time = KERNEL_time();
	LCD_cursor_data = 0;
	LCD_reverse = 0;
	next_task(LCD_main);
}

//---------------------------------------------------------------------
//又.
void LCD_main(void) {
	int c;
	uchar d;
	char *p;

	// J[\EuN
	if ((c = KERNEL_time() - LCD_cursor_time) > LCD_CURSOR_TIME) {
		LCD_cursor_time += LCD_CURSOR_TIME;
		LCD_cursor_data ^= 0xff;
		LCD_cursor(1);
	}
	
	/***************************/
	/* \&R[h  */
	/***************************/
	if ( (c = getc(&LCD_file) ) == EOF) {
		delay_task(1);
	}
	d = (uchar) c;
	if (!LCD_file.code) {
		/* 1oCgڏ */
		if (d >= 0x80)
			LCD_file.code = d;					/* 2oCgR[h1oCgڕۑ */
		else if (d >= 0x20)	{					/* p\ */
			//LCD_cursor(0);						/* J[\폜 */	/* {J[\̏ꍇ͕݂ŏ */
			LCD_font( FONT_ascii(d), -1);			/* ptHgWJ */
			LCD_cursor(1);							/* J[\\ */
		}
		else if (d == LCD_POSITION)
			LCD_file.code = d;					/* J[\ړR}hۑ */
		else LCD_control((char)d);				/* ̑R[h */
	}
	else if (LCD_file.code > 0x7f) {
		/* 2oCg\ */
		c = ((int) LCD_file.code << 8) + d;		/* 2oCgR[h̐ */
		p = FONT_kanji( (unsigned short) c);		/* 2oCgR[hSptHgʒu擾 */
		//LCD_cursor(0);						/* J[\폜 */	/* {J[\̏ꍇ͕݂ŏ */
		LCD_font(p, -1);						/* Sp̑O */
		LCD_font(p + 14, 1);					/* Sp̌㔼 */
		LCD_cursor(1);							/* J[\\ */
		LCD_file.code = 0;						/* 2oCgI */
	}
	else if (LCD_file.code == LCD_POSITION) {
		/* J[\ړ(p[^1oCgڂ̕ۑ) */
		LCD_file.type = d;
		LCD_file.code++;
	}
	else if (LCD_file.code == LCD_POSITION + 1) {
		/* J[\ړ */
		LCD_cursor(0);							/* J[\폜 */
		if (!(LCD_file.type & 2)) {				/* Έʒuړ̏ꍇ͌ʒuNA */
			LCD_v = 0;
			LCD_h = 0;
		}
		LCD_v += (schar)(LCD_file.type >> 2);
		LCD_h += (schar)(d >> 1);
		while (LCD_v >= LCD_V_MAX) LCD_v -= LCD_V_MAX;
		while (LCD_h >= LCD_H_MAX) LCD_h -= LCD_H_MAX;
		LCD_cursor(1);							/* J[\\ */
		LCD_file.code = 0;						/* ړR}hI */
	}
	retry_task();
}

//============================================================================
// [J֐.
//============================================================================
// LCD݃f[^o.
void LCD_write_data(char c) {
	
	//f[^o(݃tO̓NA).
	_regSP(SPI_SP, DATA) = c;

	//݃tO҂.
	while(!(_regSP(SPI_SP, INTFLAG) & SPI_TX_END));
		
	return;
}

//---------------------------------------------------------------------
// e퐧R[h(\b,\t,\n)̏.
void LCD_control(char c) {
	char i;

	LCD_cursor(0);					/* J[\폜 */
	switch (c) {
		case '\b':
			/*****************************************
			 obNXy[X
			 *****************************************/
			if (--LCD_h < 0) {
				LCD_h = LCD_H_MAX - 1;
				if (--LCD_v < 0) LCD_v = LCD_V_MAX - 1;
			}
			LCD_font(0, 1);
			if (--LCD_h < 0) {
				LCD_h = LCD_H_MAX - 1;
				if (--LCD_v < 0) LCD_v = LCD_V_MAX - 1;
			}
			break;
		case '\t':
			/*****************************************
			 ^u(p4P)
			 *****************************************/
			if ((i = (char) (LCD_h & 3)) ) {
				i = (char) (4 - i);
				while (i--) LCD_font(0, 1);
			}
			break;
		case '\n':
			/*****************************************
			 s
			 *****************************************/
			i = (char) (LCD_H_MAX - LCD_h);
			while (i--) {
				LCD_font(0, 1);
			}
			break;
		case LCD_REVERSE_ON:  LCD_reverse = 0xff; break;
		case LCD_REVERSE_OFF: LCD_reverse = 0;    break;
		default:
			break;
	}
	LCD_cursor(1);				/* J[\\ */
	return;
}

//---------------------------------------------------------------------
// tHg.
void LCD_font(char *p, schar i) {
	uchar j, k, m, n;

	/* ݃CʒuAJݒ */
	PORT_outclr(LCD_CS);

	m = (uchar) (0xB0 + LCD_v + LCD_v);		/* y[WAhX */
	n = (uchar) (LCD_h << 3);				/* JEAhX */

	for (k = 0; k < 2; k++) {
		/***** ݃AhXݒ *****/
		PORT_outclr(LCD_RS);				/* R}hݎw */
		LCD_write_data(m++);				/* y[WAhXݒ */
		LCD_write_data(0x10 + (n >> 4));	/* JAhXʐݒ */
		LCD_write_data(0x00 + (n & 0xf));	/* JAhXʐݒ */

		/***** tHgEf[^] *****/
		PORT_outset(LCD_RS);				/* f[^ݎw */
		if (i < 0) {						/* Oɋ󔒒ǉ */
			LCD_write_data(LCD_reverse);
		}
		
		for (j = 0; j < 7; j++) {			/* tHg] */
			if (p != NULL)	LCD_write_data((char)(*p++ ^ LCD_reverse));
			else			LCD_write_data(LCD_reverse);
		}

		if (i > 0) {						/* ɋ󔒒ǉ */
			for (j = 0; j < i; j++) {
				LCD_write_data(LCD_reverse);
			}
		}
	}

	/* LCD I */
	PORT_outset(LCD_CS);

	if (++LCD_h >= LCD_H_MAX) {
		LCD_h = 0;
		if (++LCD_v >= LCD_V_MAX) LCD_v = 0;
	}
	
	return;
}

//---------------------------------------------------------------------
// J[\\.
void LCD_cursor(char c) {
	uchar k, m, n;

	/* ݃CʒuAJݒ */
	PORT_outclr(LCD_CS);

	if (c) c = LCD_cursor_data;					/* J[\Ef[^ */
	m = (uchar) (0xB0 + LCD_v + LCD_v);			/* y[WAhX */
	n = (uchar) (LCD_h << 3);					/* JEAhX */

	for (k = 0; k < 2; k++) {
		/***** ݃AhXݒ *****/
		PORT_outclr(LCD_RS);					/* R}hݎw */
		LCD_write_data(m++);					/* y[WAhXݒ */
		LCD_write_data(0x10 + (n >> 4));	/* JAhXʐݒ */
		LCD_write_data(0x00 + (n & 0xf));	/* JAhXʐݒ */

		/***** J[\] *****/
		PORT_outset(LCD_RS);					/* f[^ݎw */
		LCD_write_data(c);
	}

	/* LCD I */
	PORT_outset(LCD_CS);

	return;
}

//============================================================================
// nh.
//============================================================================

#endif

//============================================================================
