//============================================================================
// RS485^XN
//
// (c)2021 Micon-Giken
//============================================================================
#include "kernel.h"
#include "task_rs485.h"
#include <stdio.h>

//============================================================================
// `.
//============================================================================

//============================================================================
// vg^Cv.
//============================================================================
void RS485_open_common(void);
void RS485_wake(void);

//============================================================================
// f[^.
//============================================================================
static uchar RS485_tx_buff[RS485_SEND_SIZE];	// MOEobt@.
static uchar RS485_rx_buff[RS485_RECV_SIZE];	// MOEobt@.
#if   STDIO_VERSION == 1
static FILE RS485_txf = {						// MFILE
	RS485_tx_buff, RS485_tx_buff, RS485_tx_buff, RS485_tx_buff,
	RS485_tx_buff + (RS485_SEND_SIZE - 1), 0,0,0,0,
	RS485_wake, RS485_close
};
static FILE RS485_rxf = {						// MFILE
	RS485_rx_buff, RS485_rx_buff, RS485_rx_buff, RS485_rx_buff,
	RS485_rx_buff + (RS485_RECV_SIZE - 1), 0,0,0,0,
	NULL, RS485_close
};
#elif STDIO_VERSION == 2
static FILE RS485_txf = {						// MFILE
	RS485_tx_buff, 0,0,0, RS485_SEND_SIZE - 1, 0,0,0,0,
	RS485_wake, RS485_close
};
static FILE RS485_rxf = {						// MFILE
	RS485_rx_buff, 0,0,0, RS485_RECV_SIZE - 1, 0,0,0,0,
	NULL, RS485_close
};
#endif

//============================================================================
// API֐.
//============================================================================
// RS485I[v֐
FILE* RS485_open(const char *name, const char *mode) {
	if (*mode == 'w') {
		if (RS485_txf.rw_mode)	 return NULL;
		if (RS485_rxf.rw_mode == '\0') {
			RS485_open_common();
		}
#if   STDIO_VERSION == 1
		RS485_txf.wp = RS485_tx_buff;
		RS485_txf.rp = RS485_tx_buff;
		RS485_txf.cp = RS485_tx_buff;
#elif STDIO_VERSION == 2
		RS485_txf.wp = 0;
		RS485_txf.rp = 0;
		RS485_txf.cp = 0;
#endif
		RS485_txf.rw_mode = mode[0];
		RS485_txf.binary = mode[1];
		while (_bitSC(RS485_SC, SYNCBUSY, CTRLB));
		_setSC(RS485_SC, CTRLB, TXEN, 1);
		return &RS485_txf;
	}
	else if (*mode == 'r') {
		if (RS485_rxf.rw_mode)	 return NULL;
		if (RS485_txf.rw_mode == '\0') {
			RS485_open_common();
		}
#if   STDIO_VERSION == 1
		RS485_rxf.wp = RS485_rx_buff;
		RS485_rxf.rp = RS485_rx_buff;
		RS485_rxf.cp = RS485_rx_buff;
#elif STDIO_VERSION == 2
		RS485_rxf.wp = 0;
		RS485_rxf.rp = 0;
		RS485_rxf.cp = 0;
#endif
		RS485_rxf.rw_mode = mode[0];
		RS485_rxf.binary = mode[1];
		while (_bitSC(RS485_SC, SYNCBUSY, CTRLB));
		_setSC(RS485_SC, CTRLB, RXEN, 1);
		return &RS485_rxf;
	}
	return NULL;
}

//---------------------------------------------------------------------
// RS485N[Y֐
int RS485_close(void *fp) {
	if (fp == &RS485_txf) {
		RS485_txf.rw_mode = '\0';
		while (_bitSC(RS485_SC, SYNCBUSY, CTRLB));
		_putSC(RS485_SC, CTRLB, TXEN, 0);
		if (RS485_rxf.rw_mode == '\0') {
			_putSC(RS485_SC, CTRLA, ENABLE, 0);			//~.
		}
		return 0;
	}
	else if (fp == &RS485_rxf) {
		RS485_rxf.rw_mode = '\0';
		while (_bitSC(RS485_SC, SYNCBUSY, CTRLB));
		_putSC(RS485_SC, CTRLB, RXEN, 0);
		if (RS485_txf.rw_mode == '\0') {
			_putSC(RS485_SC, CTRLA, ENABLE, 0);			//~.
		}
		return 0;
	}
	return EOF;
}

//---------------------------------------------------------------------
// MJn (ԐAPI)
void RS485_wake(void) {

#ifdef RS485_IRQ
	NVIC_disable_irq(RS485_IRQ);
	if (!(_regSC(RS485_SC, INTENSET) & RS485_TX_INT))	{
		//MX^[g
		_regSC(RS485_SC, INTENSET) = RS485_TX_INT;
	}
	NVIC_enable_irq(RS485_IRQ);
#else
	NVIC_disable_irq(RS485_DRE_IRQ);
	if (!(_regSC(RS485_SC, INTENSET) & RS485_TX_INT))	{
		//MX^[g
		_regSC(RS485_SC, INTENSET) = RS485_TX_INT;
	}
	NVIC_enable_irq(RS485_DRE_IRQ);
#endif
	return;
}

//============================================================================
// ^XNA又.
//============================================================================

//============================================================================
// [J֐.
//============================================================================
// openʏ
void RS485_open_common(void) {
	//ӗpNbNAoXNbNݒ.
	_reg(GCLK, PCHCTRL)[RS485_PCH] =
			_val(GCLK, PCHCTRL, GEN, RS485_GCLK) |
			_val(GCLK, PCHCTRL, CHEN, 1) ;
	RS485_APB_EN();										//oXNbNJn.

	//|[gݒ.
	PORT_mode_inpu(RS485_RX_PIN);
	PORT_peri(RS485_RX_PIN, RS485_MAPPING);
	PORT_peri(RS485_TX_PIN, RS485_MAPPING);
	PORT_peri(RS485_EN_PIN, RS485_MAPPING);
	
	//RS485ݒ ({[[gvZ/400́AvZrł̃I[o[t[΍).
#ifdef RS485_FREQUENCY
	_regSC(RS485_SC, BAUD) = 65536UL -	(1048576UL * (RS485_BAUDRATE / 400)) / (RS485_FREQUENCY / 400);
#else
	_regSC(RS485_SC, BAUD) = 65536UL -	(1048576UL * (RS485_BAUDRATE / 400)) / (UART_FREQUENCY / 400);
#endif
	_regSC(RS485_SC, CTRLA) =
			_valSC(RS485_SC, CTRLA, DORD, 1)	|		//LSB first.
			_valSC(RS485_SC, CTRLA, TXPO, 0x03) |		//M|[g(RS485[h).
			_valSC(RS485_SC, CTRLA, RXPO, 0x01) |		//M|[g.
			_valSC(RS485_SC, CTRLA, RXPO, 0x01) |		//M|[g.
			_valSC(RS485_SC, CTRLA, MODE, 0x01) ;		//UART[h.
//	_setSC(RS485_SC, CTRLB, COLDEN, 1);					// RWo.
	_setSC(RS485_SC, CTRLC, GTIME, 1);					// K[h(l0Ȃ̂set).
	_setSC(RS485_SC, CTRLA, ENABLE, 1);					//Jn.
	
#ifdef RS485_IRQ
	NVIC_set_priority(RS485_IRQ, 64);					//NVICDx64.
	NVIC_enable_irq(RS485_IRQ);							//NVICCl[u.
#else
	NVIC_set_priority(RS485_RXC_IRQ, 64);					//NVICDx64.
	NVIC_enable_irq(RS485_RXC_IRQ);							//NVICCl[u.
	NVIC_set_priority(RS485_DRE_IRQ, 128);					//NVICDx128.
	NVIC_enable_irq(RS485_DRE_IRQ);							//NVICCl[u.
#endif

	_regSC(RS485_SC, INTENSET) =
//			RS485_TX_INT |								//M݃Cl[u.
			RS485_RX_INT ;								//M݃Cl[u.
	return;
}

//============================================================================
// nh.
//============================================================================
#ifdef RS485_IRQ
__attribute__((interrupt))
void RS485_Handler(void) {
	
	if (_regSC(RS485_SC, INTFLAG) & RS485_RX_INT) {
		_regSC(RS485_SC, INTFLAG) = RS485_RX_INT;
		if (_bitSC(RS485_SC, STATUS, FERR)) {
			_regSC(RS485_SC, STATUS) = _valSC(RS485_SC, STATUS, FERR, 1);
#if   STDIO_VERSION == 1
			*RS485_rxf.wp = _regSC(RS485_SC, DATA);
#elif STDIO_VERSION == 2
			RS485_rxf.buff[RS485_rxf.wp] = _regSC(RS485_SC, DATA);
#endif
		}
		else {
#if   STDIO_VERSION == 1
			*RS485_rxf.wp = _regSC(RS485_SC, DATA);
			if (RS485_rxf.wp < RS485_rxf.end) RS485_rxf.wp++; else RS485_rxf.wp = RS485_rxf.begin;
#elif STDIO_VERSION == 2
			RS485_rxf.buff[RS485_rxf.wp++] = _regSC(RS485_SC, DATA);
#endif
		}
	}
	else if (_regSC(RS485_SC, INTFLAG) & RS485_TX_INT) {
		_regSC(RS485_SC, INTFLAG) = RS485_TX_INT;
		if (RS485_txf.rp != RS485_txf.wp) {
#if   STDIO_VERSION == 1
			*(char*)&(_regSC(RS485_SC, DATA)) = *RS485_txf.rp;
			if (RS485_txf.rp < RS485_txf.end) RS485_txf.rp++; else RS485_txf.rp = RS485_txf.begin;
#elif STDIO_VERSION == 2
			*(char*)&(_regSC(RS485_SC, DATA)) = RS485_txf.buff[RS485_txf.rp++];
#endif
		}
		else {
			_regSC(RS485_SC, INTENCLR) = RS485_TX_INT;
		}
	}
}

#else
__attribute__((interrupt))
void RS485_RXC_Handler(void) {
	_regSC(RS485_SC, INTFLAG) = RS485_RX_INT;
	if (_bitSC(RS485_SC, STATUS, FERR)) {
		_regSC(RS485_SC, STATUS) = _valSC(RS485_SC, STATUS, FERR, 1);
#if   STDIO_VERSION == 1
		*RS485_rxf.wp = _regSC(RS485_SC, DATA);
#elif STDIO_VERSION == 2
		RS485_rxf.buff[RS485_rxf.wp] = _regSC(RS485_SC, DATA);
#endif
	}
	else {
#if   STDIO_VERSION == 1
		*RS485_rxf.wp = _regSC(RS485_SC, DATA);
		if (RS485_rxf.wp < RS485_rxf.end) RS485_rxf.wp++; else RS485_rxf.wp = RS485_rxf.begin;
#elif STDIO_VERSION == 2
		RS485_rxf.buff[RS485_rxf.wp++] = _regSC(RS485_SC, DATA);
#endif
	}
}

__attribute__((interrupt))
void RS485_DRE_Handler(void) {
	_regSC(RS485_SC, INTFLAG) = RS485_TX_INT;
	if (RS485_txf.rp != RS485_txf.wp) {
#if   STDIO_VERSION == 1
		*(char*)&(_regSC(RS485_SC, DATA)) = *RS485_txf.rp;
		if (RS485_txf.rp < RS485_txf.end) RS485_txf.rp++; else RS485_txf.rp = RS485_txf.begin;
#elif STDIO_VERSION == 2
		*(char*)&(_regSC(RS485_SC, DATA)) = RS485_txf.buff[RS485_txf.rp++];
#endif
	}
	else {
		_regSC(RS485_SC, INTENCLR) = RS485_TX_INT;
	}
}
#endif

//============================================================================
