//============================================================================
// UART^XN
//
// (c)2021 Micon-Giken
//============================================================================
#include "kernel.h"
#include "task_uart.h"
#include <stdio.h>

//============================================================================
// `.
//============================================================================

//============================================================================
// vg^Cv.
//============================================================================
void UART_open_common(void);
void UART_wake(void);

//============================================================================
// f[^.
//============================================================================
static uchar UART_tx_buff[UART_SEND_SIZE];	// MOEobt@.
static uchar UART_rx_buff[UART_RECV_SIZE];	// MOEobt@.

#if   STDIO_VERSION == 1
static FILE UART_txf = {					// MFILE
	UART_tx_buff, UART_tx_buff, UART_tx_buff, UART_tx_buff,
	UART_tx_buff + (UART_SEND_SIZE - 1), 0,0,0,0,
	UART_wake, UART_close
};
static FILE UART_rxf = {					// MFILE
	UART_rx_buff, UART_rx_buff, UART_rx_buff, UART_rx_buff,
	UART_rx_buff + (UART_RECV_SIZE - 1), 0,0,0,0,
	NULL, UART_close
};
#elif STDIO_VERSION == 2
static FILE UART_txf = {					// MFILE
	UART_tx_buff, 0,0,0, UART_SEND_SIZE - 1, 0,0,0,0,
	UART_wake, UART_close
};
static FILE UART_rxf = {					// MFILE
	UART_rx_buff, 0,0,0, UART_RECV_SIZE - 1, 0,0,0,0,
	NULL, UART_close
};
#endif

//============================================================================
// API֐.
//============================================================================
// UARTI[v֐
FILE* UART_open(const char *name, const char *mode) {
	if (*mode == 'w') {
		if (UART_txf.rw_mode)	return NULL;
		if (UART_rxf.rw_mode == '\0') {
			UART_open_common();
		}
#if   STDIO_VERSION == 1
		UART_txf.wp = UART_tx_buff;
		UART_txf.rp = UART_tx_buff;
		UART_txf.cp = UART_tx_buff;
#elif STDIO_VERSION == 2
		UART_txf.wp = 0;
		UART_txf.rp = 0;
		UART_txf.cp = 0;
#endif
		UART_txf.rw_mode = mode[0];
		UART_txf.binary = mode[1];
		while (_bitSC(UART_SC, SYNCBUSY, CTRLB));
		_setSC(UART_SC, CTRLB, TXEN, 1);
		return &UART_txf;
	}
	else if (*mode == 'r') {
		if (UART_rxf.rw_mode)	return NULL;
		if (UART_txf.rw_mode == '\0') {
			UART_open_common();
		}
#if   STDIO_VERSION == 1
		UART_rxf.wp = UART_rx_buff;
		UART_rxf.rp = UART_rx_buff;
		UART_rxf.cp = UART_rx_buff;
#elif STDIO_VERSION == 2
		UART_rxf.wp = 0;
		UART_rxf.rp = 0;
		UART_rxf.cp = 0;
#endif
		UART_rxf.rw_mode = mode[0];
		UART_rxf.binary = mode[1];
		while (_bitSC(UART_SC, SYNCBUSY, CTRLB));
		_setSC(UART_SC, CTRLB, RXEN, 1);
		return &UART_rxf;
	}
	return NULL;
}

//---------------------------------------------------------------------
// UARTN[Y֐
int UART_close(void *fp) {
	if (fp == &UART_txf) {
		UART_txf.rw_mode = '\0';
		while (_bitSC(UART_SC, SYNCBUSY, CTRLB));
		_putSC(UART_SC, CTRLB, TXEN, 0);
		if (UART_rxf.rw_mode == '\0') {
			_putSC(UART_SC, CTRLA, ENABLE, 0);		//~.
		}
		return 0;
	}
	else if (fp == &UART_rxf) {
		UART_rxf.rw_mode = '\0';
		while (_bitSC(UART_SC, SYNCBUSY, CTRLB));
		_putSC(UART_SC, CTRLB, RXEN, 0);
		if (UART_txf.rw_mode == '\0') {
			_putSC(UART_SC, CTRLA, ENABLE, 0);		//~.
		}
		return 0;
	}
	return EOF;
}

//---------------------------------------------------------------------
// MJn (ԐAPI)
void UART_wake(void) {
	
#ifdef UART_IRQ
	NVIC_disable_irq(UART_IRQ);
	if (!(_regSC(UART_SC, INTENSET) & UART_TX_INT))	{
		//MX^[g
		_regSC(UART_SC, INTENSET) = UART_TX_INT;
	}
	NVIC_enable_irq(UART_IRQ);
#else
	NVIC_disable_irq(UART_DRE_IRQ);
	if (!(_regSC(UART_SC, INTENSET) & UART_TX_INT))	{
		//MX^[g
		_regSC(UART_SC, INTENSET) = UART_TX_INT;
	}
	NVIC_enable_irq(UART_DRE_IRQ);
#endif
	return;
}

//============================================================================
// ^XNA又.
//============================================================================

//============================================================================
// [J֐.
//============================================================================
// openʏ
void UART_open_common(void) {
	//ӗpNbNAoXNbNݒ.
#if (TARGET_BOARD & SAMD21_ALL)
	_reg(GCLK, CLKCTRL) =
			_val(GCLK, CLKCTRL, ID, UART_PCH) |
			_val(GCLK, CLKCTRL, GEN, UART_GCLK) |
			_val(GCLK, CLKCTRL, CLKEN, 1) ;
#else
	_reg(GCLK, PCHCTRL)[UART_PCH] =
			_val(GCLK, PCHCTRL, GEN, UART_GCLK) |
			_val(GCLK, PCHCTRL, CHEN, 1) ;
#endif
	UART_APB_EN();										//oXNbNJn.
	
	//|[gݒ.
	PORT_mode_inpu(UART_RX_PIN);
	PORT_peri(UART_RX_PIN, UART_MAPPING);
	PORT_peri(UART_TX_PIN, UART_MAPPING);
	
	//UARTݒ ({[[gvZ/400́AvZrł̃I[o[t[΍).
	_regSC(UART_SC, BAUD) = 65536UL -  (1048576UL * (UART_BAUDRATE / 400)) / (UART_FREQUENCY / 400);
	_regSC(UART_SC, CTRLA) =
			_valSC(UART_SC, CTRLA, DORD, 1)	   |		//LSB first.
			_valSC(UART_SC, CTRLA, TXPO, 0x00) |		//M|[g.
			_valSC(UART_SC, CTRLA, RXPO, 0x01) |		//M|[g.
			_valSC(UART_SC, CTRLA, MODE, 0x01) ;		//UART[h.
	_setSC(UART_SC, CTRLA, ENABLE, 1);					//Jn.

#ifdef UART_IRQ
	NVIC_set_priority(UART_IRQ, 64);					//NVICDx64.
	NVIC_enable_irq(UART_IRQ);							//NVICCl[u.
#else
	NVIC_set_priority(UART_RXC_IRQ, 64);					//NVICDx64.
	NVIC_enable_irq(UART_RXC_IRQ);							//NVICCl[u.
	NVIC_set_priority(UART_DRE_IRQ, 128);					//NVICDx128.
	NVIC_enable_irq(UART_DRE_IRQ);							//NVICCl[u.
#endif

	_regSC(UART_SC, INTENSET) =
//			UART_TX_INT |								//M݃Cl[u.
			UART_RX_INT ;								//M݃Cl[u.
	return;
}

//============================================================================
// nh.
//============================================================================
#ifdef UART_Handler
__attribute__((interrupt))
void UART_Handler(void) {

	if (_regSC(UART_SC, INTFLAG) & UART_RX_INT) {
		_regSC(UART_SC, INTFLAG) = UART_RX_INT;
		if (_bitSC(UART_SC, STATUS, FERR)) {
			_regSC(UART_SC, STATUS) = _valSC(UART_SC, STATUS, FERR, 1);
#if	  STDIO_VERSION == 1
			*UART_rxf.wp = _regSC(UART_SC, DATA);
#elif STDIO_VERSION == 2
			UART_rxf.buff[UART_rxf.wp] = _regSC(UART_SC, DATA);
#endif
		}
		else {
#if	  STDIO_VERSION == 1
			*UART_rxf.wp = _regSC(UART_SC, DATA);
			if (UART_rxf.wp < UART_rxf.end) UART_rxf.wp++; else UART_rxf.wp = UART_rxf.begin;
#elif STDIO_VERSION == 2
			UART_rxf.buff[UART_rxf.wp++] = _regSC(UART_SC, DATA);
#endif
		}
	}
	else if (_regSC(UART_SC, INTFLAG) & UART_TX_INT) {
		_regSC(UART_SC, INTFLAG) = UART_TX_INT;
		if (UART_txf.rp != UART_txf.wp) {
#if	  STDIO_VERSION == 1
			*(char*)&(_regSC(UART_SC, DATA)) = *UART_txf.rp;
			if (UART_txf.rp < UART_txf.end) UART_txf.rp++; else UART_txf.rp = UART_txf.begin;
#elif STDIO_VERSION == 2
			*(char*)&(_regSC(UART_SC, DATA)) = UART_txf.buff[UART_txf.rp++];
#endif
		}
		else {
			_regSC(UART_SC, INTENCLR) = UART_TX_INT;
		}
	}
}
#else
__attribute__((interrupt))
void UART_RXC_Handler(void) {

	_regSC(UART_SC, INTFLAG) = UART_RX_INT;
	if (_bitSC(UART_SC, STATUS, FERR)) {
		_regSC(UART_SC, STATUS) = _valSC(UART_SC, STATUS, FERR, 1);
#if	  STDIO_VERSION == 1
		*UART_rxf.wp = _regSC(UART_SC, DATA);
#elif STDIO_VERSION == 2
		UART_rxf.buff[UART_rxf.wp] = _regSC(UART_SC, DATA);
#endif
	}
	else {
#if	  STDIO_VERSION == 1
		*UART_rxf.wp = _regSC(UART_SC, DATA);
		if (UART_rxf.wp < UART_rxf.end) UART_rxf.wp++; else UART_rxf.wp = UART_rxf.begin;
#elif STDIO_VERSION == 2
		UART_rxf.buff[UART_rxf.wp++] = _regSC(UART_SC, DATA);
#endif
	}
}

__attribute__((interrupt))
void UART_DRE_Handler(void) {

	_regSC(UART_SC, INTFLAG) = UART_TX_INT;
	if (UART_txf.rp != UART_txf.wp) {
#if	  STDIO_VERSION == 1
		*(char*)&(_regSC(UART_SC, DATA)) = *UART_txf.rp;
		if (UART_txf.rp < UART_txf.end) UART_txf.rp++; else UART_txf.rp = UART_txf.begin;
#elif STDIO_VERSION == 2
		*(char*)&(_regSC(UART_SC, DATA)) = UART_txf.buff[UART_txf.rp++];
#endif
	}
	else {
		_regSC(UART_SC, INTENCLR) = UART_TX_INT;
	}
}
#endif

//============================================================================
