//============================================================================
// USB^XN
//
// (c)2021 Micon-Giken
//============================================================================
#include "kernel.h"
#include "task_usb.h"
#include <stdio.h>

//============================================================================
// `.
//============================================================================
//#define APP_READ_BUFFER_SIZE			512
#define APP_READ_BUFFER_SIZE			256		// 256łǂA[hTCYς̂NĜ悤.

typedef struct {
    /* Device layer handle returned by device layer open function */
    USB_DEVICE_HANDLE deviceHandle;
    /* Set Line Coding Data */
    USB_CDC_LINE_CODING setLineCodingData;
    /* Device configured state */
    bool isConfigured;
    /* Get Line Coding Data */
    USB_CDC_LINE_CODING getLineCodingData;
    /* Control Line State */
    USB_CDC_CONTROL_LINE_STATE controlLineStateData;
    /* Read transfer handle */
    USB_DEVICE_CDC_TRANSFER_HANDLE readTransferHandle;
    /* Write transfer handle */
    USB_DEVICE_CDC_TRANSFER_HANDLE writeTransferHandle;
    /* True if a character was read */
    bool	isReadComplete;
    /* True if a character was written*/
    bool isWriteComplete;
	/* read write counter*/
	int		nReadComplete;
    int		nWriteComplete;
    /* Flag determines SOF event occurrence */
    bool sofEventHasOccurred;
    /* Break data */
    uint16_t breakData;
    /* Application CDC read buffer */
    uint8_t * cdcReadBuffer;
    /* Application CDC Write buffer */
    uint8_t * cdcWriteBuffer;
    /* Number of bytes read from Host */ 
    uint32_t numBytesRead; 
} APP_DATA;

//============================================================================
// vg^Cv.
//============================================================================
void USB_wake(void);
void USB_task_init(void);			/* initial state. */
void USB_task_conf(void);			/* Waiting for device configuration*/
void USB_task_read(void);			/* read & write */
void USB_task_error(void);		/* Error state*/

DRV_USB_VBUS_LEVEL DRV_USBFSV1_VBUS_Comparator(void);
bool APP_StateReset(void);
void APP_USBDeviceEventHandler (USB_DEVICE_EVENT event, void * eventData, uintptr_t context);

//============================================================================
// f[^.
//============================================================================
static uchar USB_tx_buff[USB_SEND_SIZE];	// MOEobt@.
static uchar USB_rx_buff[USB_RECV_SIZE];	// MOEobt@.
#if   STDIO_VERSION == 1
static FILE USB_txf = {					// MFILE
	USB_tx_buff, USB_tx_buff, USB_tx_buff, USB_tx_buff,
	USB_tx_buff + (USB_SEND_SIZE - 1), 0,0,0,0,
	USB_wake, USB_close
};
static FILE USB_rxf = {					// MFILE
	USB_rx_buff, USB_rx_buff, USB_rx_buff, USB_rx_buff,
	USB_rx_buff + (USB_RECV_SIZE - 1), 0,0,0,0,
	NULL, USB_close
};
#elif STDIO_VERSION == 2
static FILE USB_txf = {					// MFILE
	USB_tx_buff, 0,0,0, USB_SEND_SIZE - 1, 0,0,0,0,
	USB_wake, USB_close
};
static FILE USB_rxf = {					// MFILE
	USB_rx_buff, 0,0,0, USB_RECV_SIZE - 1, 0,0,0,0,
	NULL, USB_close
};
#endif

SYSTEM_OBJECTS sysObj;
uint8_t	cdcReadBuffer[APP_READ_BUFFER_SIZE];
uint8_t	cdcWriteBuffer[USB_SEND_SIZE];
//uint8_t	cdcWriteBuffer[APP_READ_BUFFER_SIZE];
APP_DATA appData;

const DRV_USBFSV1_INIT drvUSBInit = {
	/* Interrupt Source for USB module */
	.interruptSource = USB_IRQ,
#if (TARGET_BOARD & SAME54_ALL)
	/* Interrupt Source for USB module */
	.interruptSource1 = USB2_IRQ,
	/* Interrupt Source for USB module */
	.interruptSource2 = USB3_IRQ,
	/* Interrupt Source for USB module */
	.interruptSource3 = USB4_IRQ,
#endif
	/* System module initialization */
	.moduleInit = {0},
	/* USB Controller to operate as USB Device */
	.operationMode = DRV_USBFSV1_OPMODE_DEVICE,
	/* USB Full Speed Operation */
	.operationSpeed = USB_SPEED_FULL,
    /* Stop in idle */
	.runInStandby = true,
	/* Suspend in sleep */
	.suspendInSleep = false,
	/* Identifies peripheral (PLIB-level) ID */
	.usbID = USB_REGS,
	/* Function to check for VBus */
	.vbusComparator = DRV_USBFSV1_VBUS_Comparator
};

//============================================================================
// API֐.
//============================================================================
// UARTI[v֐.
FILE* USB_open(const char *name, const char *mode) {
	if (*mode == 'w') {
		if (USB_txf.rw_mode)	return NULL;
#if   STDIO_VERSION == 1
		USB_txf.wp = USB_tx_buff;
		USB_txf.rp = USB_tx_buff;
		USB_txf.cp = USB_tx_buff;
#elif STDIO_VERSION == 2
		USB_txf.wp = 0;
		USB_txf.rp = 0;
		USB_txf.cp = 0;
#endif
		USB_txf.rw_mode = mode[0];
		USB_txf.binary = mode[1];
		return &USB_txf;
	}
	else if (*mode == 'r') {
		if (USB_rxf.rw_mode)	return NULL;
#if   STDIO_VERSION == 1
		USB_rxf.wp = USB_rx_buff;
		USB_rxf.rp = USB_rx_buff;
		USB_rxf.cp = USB_rx_buff;
#elif STDIO_VERSION == 2
		USB_rxf.wp = 0;
		USB_rxf.rp = 0;
		USB_rxf.cp = 0;
#endif
		USB_rxf.rw_mode = mode[0];
		USB_rxf.binary = mode[1];
		return &USB_rxf;
	}
	return NULL;
}

//---------------------------------------------------------------------
// UARTN[Y֐
int USB_close(void *fp) {
	if (fp == &USB_txf) {
		USB_txf.rw_mode = '\0';
		return 0;
	}
	else if (fp == &USB_rxf) {
		USB_rxf.rw_mode = '\0';
		return 0;
	}
	return EOF;
}

//---------------------------------------------------------------------
// MJn (ԐAPI)
void USB_wake(void) {
	wakeup_task(USB_taskid);
	return;
}

//---------------------------------------------------------------------
// USB(CDC)hCoE^XN.
void USB_init(void) {
	//ӗpNbNAoXNbNݒ.
#if (TARGET_BOARD & SAMD21_ALL)
	_reg(GCLK, CLKCTRL) =
			_val(GCLK, CLKCTRL, ID, USB_PCH) |
			_val(GCLK, CLKCTRL, GEN, USB_GCLK) |
			_val(GCLK, CLKCTRL, CLKEN, 1) ;
#else
	_reg(GCLK, PCHCTRL)[USB_PCH] =
			_val(GCLK, PCHCTRL, GEN, USB_GCLK) |
			_val(GCLK, PCHCTRL, CHEN, 1) ;
#endif
#if (TARGET_BOARD & SAME54_ALL)
	USB_APB_EN();
#endif
	USB_AHB_EN();										//oXNbNJn.

	// |[gݒ.
	PORT_mode_in(USB_VBUS_PIN);
	PORT_peri(USB_DP_PIN, USB_MAPPING);
	PORT_peri(USB_DM_PIN, USB_MAPPING);
	
	/* Initialize USB Driver */ 
    sysObj.drvUSBFSV1Object = DRV_USBFSV1_Initialize(DRV_USBFSV1_INDEX_0, (SYS_MODULE_INIT *) &drvUSBInit);	

	/* Initialize the USB device layer */
    sysObj.usbDevObject0 = USB_DEVICE_Initialize (USB_DEVICE_INDEX_0 , ( SYS_MODULE_INIT* ) & usbDevInitData);

	//------------------------------------
	// CDC.
	//------------------------------------
    /* Device Layer Handle  */
    appData.deviceHandle = USB_DEVICE_HANDLE_INVALID ;
    /* Device configured status */
    appData.isConfigured = false;
    /* Initial get line coding state */
    appData.getLineCodingData.dwDTERate = USB_BAUDRATE;
    appData.getLineCodingData.bParityType = 0;
    appData.getLineCodingData.bParityType = 0;
    appData.getLineCodingData.bDataBits = 8;
    /* Read Transfer Handle */
    appData.readTransferHandle = USB_DEVICE_CDC_TRANSFER_HANDLE_INVALID;
    /* Write Transfer Handle */
    appData.writeTransferHandle = USB_DEVICE_CDC_TRANSFER_HANDLE_INVALID;
    /* Initialize the read complete flag */
    appData.isReadComplete = true;
	appData.nReadComplete = EOF;
    /*Initialize the write complete flag*/
    appData.isWriteComplete = true;
    /* Reset other flags */
    appData.sofEventHasOccurred = false;
    /* Set up the read buffer */
    appData.cdcReadBuffer = &cdcReadBuffer[0];
    /* Set up the read buffer */
    appData.cdcWriteBuffer = &cdcWriteBuffer[0];       

	// ݐݒ.
	NVIC_set_priority(USB_IRQ, 128);					//NVICDx128.
	NVIC_enable_irq(USB_IRQ);							//NVICCl[u.
#if (TARGET_BOARD & SAME54_ALL)
	NVIC_set_priority(USB2_IRQ, 128);					//NVICDx128.
	NVIC_enable_irq(USB2_IRQ);							//NVICCl[u.
	NVIC_set_priority(USB3_IRQ, 128);					//NVICDx128.
	NVIC_enable_irq(USB3_IRQ);							//NVICCl[u.
	NVIC_set_priority(USB4_IRQ, 128);					//NVICDx128.
	NVIC_enable_irq(USB4_IRQ);							//NVICCl[u.
#endif

	//---------------
	//^XNN
	//---------------
	start_task(USB_taskid, USB_task_init);
}

//============================================================================
// ^XNA又.
//============================================================================
//----------------------------------------------------
//ԑJ.
//----------------------------------------------------
/* Application's state machine's initial state. */
void USB_task_init(void) {
    DRV_USBFSV1_Tasks(sysObj.drvUSBFSV1Object);	/* USB FS Driver Task Routine */ 
    USB_DEVICE_Tasks(sysObj.usbDevObject0);		/* USB Device layer tasks routine */ 

	/* Open the device layer */
	appData.deviceHandle = USB_DEVICE_Open( USB_DEVICE_INDEX_0, DRV_IO_INTENT_READWRITE );

	if(appData.deviceHandle != USB_DEVICE_HANDLE_INVALID) {
		/* Register a callback with device layer to get event notification (for end point 0) */
		USB_DEVICE_EventHandlerSet(appData.deviceHandle, APP_USBDeviceEventHandler, 0);
		set_retry_task();
		next_task(USB_task_conf);
	}
	else {
		/* The Device Layer is not ready to be opened. We should try
		 * again later. */
		delay_task(5);
	}
}

//----------------------------------------------------
/* Application waits for device configuration*/
void USB_task_conf(void) {
    DRV_USBFSV1_Tasks(sysObj.drvUSBFSV1Object);	/* USB FS Driver Task Routine */ 
    USB_DEVICE_Tasks(sysObj.usbDevObject0);		/* USB Device layer tasks routine */ 

	/* Check if the device was configured */
	if(appData.isConfigured) {
		/* If the device is configured then lets start reading */
		set_retry_task();
		next_task(USB_task_read);
	}
	delay_task(5);
}

//----------------------------------------------------
/* The application checks if a switch was pressed */
void USB_task_read(void) {
    DRV_USBFSV1_Tasks(sysObj.drvUSBFSV1Object);	/* USB FS Driver Task Routine */ 
    USB_DEVICE_Tasks(sysObj.usbDevObject0);		/* USB Device layer tasks routine */ 
	if(APP_StateReset()) { next_task(USB_task_conf); }

#if   STDIO_VERSION == 1
	if (appData.nReadComplete != EOF) {
		/*݂̕ǂ.obt@~USBs.*/
		while (appData.nReadComplete < appData.numBytesRead) {
			*USB_rxf.wp = appData.cdcReadBuffer[appData.nReadComplete++];
			if (USB_rxf.wp < USB_rxf.end) USB_rxf.wp++; else USB_rxf.wp = USB_rxf.begin;
		}
		appData.nReadComplete = EOF;
	}
#elif STDIO_VERSION == 2
	if (appData.nReadComplete != EOF) {
		/*݂̕ǂ.obt@~USBs.*/
		while (appData.nReadComplete < appData.numBytesRead) {
			USB_rxf.buff[USB_rxf.wp++] = appData.cdcReadBuffer[appData.nReadComplete++];
		}
		appData.nReadComplete = EOF;
	}
#endif
	if (appData.nReadComplete == EOF) {
		if (appData.isReadComplete == true) {
			appData.isReadComplete = false;
			appData.readTransferHandle =  USB_DEVICE_CDC_TRANSFER_HANDLE_INVALID;
			USB_DEVICE_CDC_Read (USB_DEVICE_CDC_INDEX_0,
					&appData.readTransferHandle, appData.cdcReadBuffer,
					APP_READ_BUFFER_SIZE);
			if(appData.readTransferHandle == USB_DEVICE_CDC_TRANSFER_HANDLE_INVALID) {
				next_task(USB_task_error);
			}
		}
	}
	if (USB_txf.wp != USB_txf.rp) {
		if (appData.isWriteComplete == true) {
			appData.nWriteComplete = 0;
			while(USB_txf.wp != USB_txf.rp) {
#if   STDIO_VERSION == 1
				appData.cdcWriteBuffer[appData.nWriteComplete++] = *USB_txf.rp;
				if (USB_txf.rp < USB_txf.end) USB_txf.rp++; else USB_txf.rp = USB_txf.begin;
#elif STDIO_VERSION == 2
				appData.cdcWriteBuffer[appData.nWriteComplete++] = USB_txf.buff[USB_txf.rp++];
#endif
			}
            appData.writeTransferHandle = USB_DEVICE_CDC_TRANSFER_HANDLE_INVALID;
            appData.isWriteComplete = false;
			USB_DEVICE_CDC_Write(USB_DEVICE_CDC_INDEX_0,
					&appData.writeTransferHandle,
					appData.cdcWriteBuffer, appData.nWriteComplete,
					USB_DEVICE_CDC_TRANSFER_FLAGS_DATA_COMPLETE);
		}
		else {
			retry_task();
		}
	}
	delay_task(5);
}

//----------------------------------------------------
/* Application Error state*/
void USB_task_error(void) {
//	delay_task(1000);
}

//============================================================================
// [J֐.
//============================================================================
// VBUSo.
DRV_USB_VBUS_LEVEL DRV_USBFSV1_VBUS_Comparator(void) {
    DRV_USB_VBUS_LEVEL retVal = DRV_USB_VBUS_LEVEL_INVALID;
    if(true == USB_VBUS_SENSE_Get()) {
        retVal = DRV_USB_VBUS_LEVEL_VALID;
    }
	return (retVal);
}

// USBZbg.
bool APP_StateReset(void) {
    bool retVal;

    if(appData.isConfigured == false) {
        appData.readTransferHandle = USB_DEVICE_CDC_TRANSFER_HANDLE_INVALID;
        appData.writeTransferHandle = USB_DEVICE_CDC_TRANSFER_HANDLE_INVALID;
        appData.isReadComplete = true;
		appData.nReadComplete = EOF;
        appData.isWriteComplete = true;
        retVal = true;
    }
    else {
        retVal = false;
    }

    return(retVal);
}

//============================================================================
// nh.
//============================================================================
__attribute__((interrupt))
void USB_Handler(void) {
	__EI();		//d݋
	DRV_USBFSV1_Tasks_ISR(DRV_USBFSV1_INDEX_0);
}

#if (TARGET_BOARD & SAME54_ALL)
__attribute__((interrupt))
void USB2_Handler(void) {
	__EI();		//d݋
	DRV_USBFSV1_Tasks_ISR(DRV_USBFSV1_INDEX_0);
}

__attribute__((interrupt))
void USB3_Handler(void) {
	__EI();		//d݋
	DRV_USBFSV1_Tasks_ISR(DRV_USBFSV1_INDEX_0);
}

__attribute__((interrupt))
void USB4_Handler(void) {
	__EI();		//d݋
	DRV_USBFSV1_Tasks_ISR(DRV_USBFSV1_INDEX_0);
}
#endif

/*******************************************************
 * USB CDC Device Events - Application Event Handler
 *******************************************************/
USB_DEVICE_CDC_EVENT_RESPONSE APP_USBDeviceCDCEventHandler(USB_DEVICE_CDC_INDEX index,
    USB_DEVICE_CDC_EVENT event, void * pData, uintptr_t userData) {
    APP_DATA * appDataObject;
    USB_CDC_CONTROL_LINE_STATE * controlLineStateData;
    USB_DEVICE_CDC_EVENT_DATA_READ_COMPLETE * eventDataRead;
    
    appDataObject = (APP_DATA *)userData;
    switch(event) {
        case USB_DEVICE_CDC_EVENT_GET_LINE_CODING:
            /* This means the host wants to know the current line
             * coding. This is a control transfer request. Use the
             * USB_DEVICE_ControlSend() function to send the data to
             * host.  */
            USB_DEVICE_ControlSend(appDataObject->deviceHandle,
                    &appDataObject->getLineCodingData, sizeof(USB_CDC_LINE_CODING));
            break;

        case USB_DEVICE_CDC_EVENT_SET_LINE_CODING:
            /* This means the host wants to set the line coding.
             * This is a control transfer request. Use the
             * USB_DEVICE_ControlReceive() function to receive the
             * data from the host */
            USB_DEVICE_ControlReceive(appDataObject->deviceHandle,
                    &appDataObject->setLineCodingData, sizeof(USB_CDC_LINE_CODING));
            break;

        case USB_DEVICE_CDC_EVENT_SET_CONTROL_LINE_STATE:
            /* This means the host is setting the control line state.
             * Read the control line state. We will accept this request
             * for now. */
            controlLineStateData = (USB_CDC_CONTROL_LINE_STATE *)pData;
            appDataObject->controlLineStateData.dtr = controlLineStateData->dtr;
            appDataObject->controlLineStateData.carrier = controlLineStateData->carrier;
            USB_DEVICE_ControlStatus(appDataObject->deviceHandle, USB_DEVICE_CONTROL_STATUS_OK);
            break;

        case USB_DEVICE_CDC_EVENT_SEND_BREAK:
            /* This means that the host is requesting that a break of the
             * specified duration be sent. Read the break duration */
            appDataObject->breakData = ((USB_DEVICE_CDC_EVENT_DATA_SEND_BREAK *)pData)->breakDuration;
            /* Complete the control transfer by sending a ZLP  */
            USB_DEVICE_ControlStatus(appDataObject->deviceHandle, USB_DEVICE_CONTROL_STATUS_OK);
            break;

        case USB_DEVICE_CDC_EVENT_READ_COMPLETE:
            /* This means that the host has sent some data*/
            eventDataRead = (USB_DEVICE_CDC_EVENT_DATA_READ_COMPLETE *)pData;
            if(eventDataRead->status != USB_DEVICE_CDC_RESULT_ERROR) {
                appDataObject->isReadComplete = true;
                appDataObject->nReadComplete = 0;
                appDataObject->numBytesRead = eventDataRead->length; 
//				wakeup_task(USB_taskid);		//delay_taskƂ̕p֎~.
            }
            break;

        case USB_DEVICE_CDC_EVENT_CONTROL_TRANSFER_DATA_RECEIVED:
            /* The data stage of the last control transfer is
             * complete. For now we accept all the data */
            USB_DEVICE_ControlStatus(appDataObject->deviceHandle, USB_DEVICE_CONTROL_STATUS_OK);
            break;

        case USB_DEVICE_CDC_EVENT_CONTROL_TRANSFER_DATA_SENT:
            /* This means the GET LINE CODING function data is valid. We don't
             * do much with this data in this demo. */
            break;

        case USB_DEVICE_CDC_EVENT_WRITE_COMPLETE:
            /* This means that the data write got completed. We can schedule
             * the next read. */
            appDataObject->isWriteComplete = true;
//			wakeup_task(USB_taskid);		//delay_taskƂ̕p֎~.
            break;

        default:
            break;
    }

    return USB_DEVICE_CDC_EVENT_RESPONSE_NONE;
}

/***********************************************
 * Application USB Device Layer Event Handler.
 ***********************************************/
void APP_USBDeviceEventHandler (USB_DEVICE_EVENT event, void * eventData, uintptr_t context) {
    USB_DEVICE_EVENT_DATA_CONFIGURED *configuredEventData;

    switch(event) {
        case USB_DEVICE_EVENT_SOF:			//:XCb`ŎgĂ̂ŕsvH.
            /* This event is used for switch debounce. This flag is reset
             * by the switch process routine. */
            appData.sofEventHasOccurred = true;
            break;

        case USB_DEVICE_EVENT_RESET:
            appData.isConfigured = false;
            break;

        case USB_DEVICE_EVENT_CONFIGURED:
            /* Check the configuration. We only support configuration 1 */
            configuredEventData = (USB_DEVICE_EVENT_DATA_CONFIGURED*)eventData;
            
            if ( configuredEventData->configurationValue == 1) {
                /* Register the CDC Device application event handler here.
                 * Note how the appData object pointer is passed as the
                 * user data */
                USB_DEVICE_CDC_EventHandlerSet(USB_DEVICE_CDC_INDEX_0, APP_USBDeviceCDCEventHandler, (uintptr_t)&appData);

                /* Mark that the device is now configured */
                appData.isConfigured = true;
            }
            break;

        case USB_DEVICE_EVENT_POWER_DETECTED:
            /* VBUS was detected. We can attach the device */
            USB_DEVICE_Attach(appData.deviceHandle);
            break;

        case USB_DEVICE_EVENT_POWER_REMOVED:
            /* VBUS is not available. We can detach the device */
            USB_DEVICE_Detach(appData.deviceHandle);
            appData.isConfigured = false;
            break;

        case USB_DEVICE_EVENT_SUSPENDED:
            break;

        case USB_DEVICE_EVENT_RESUMED:
        case USB_DEVICE_EVENT_ERROR:
        default:
            break;
    }
}

//============================================================================
